import random


def sign(x):
    if x > 0:
        return 1
    elif x < 0:
        return -1
    return 0


class Point:
    def __init__(self, x: int, y: int):
        self.x = x
        self.y = y

    def __repr__(self):
        return f"(x={self.x:2d} ; y={self.y:2d})"

    def __add__(self, p):
        return Point(self.x + p.x, self.y + p.y)

    def __sub__(self, p):
        return Point(self.x - p.x, self.y - p.y)

    def __eq__(self, p):
        if self.x == p.x and self.y == p.y:
            return True
        return False

    def sign(self):
        return Point(sign(self.x), sign(self.y))

    def norm2(self):
        return self.x ** 2 + self.y ** 2


class Entity:
    def __init__(self, x: int, y: int, name: str):
        self.pos = Point(x, y)
        self.last_pos = Point(-1, -1)
        self.name = name

    def __repr__(self):
        return f"{self.name}{self.pos}"

    def move(self, xmin, ymin, xmax, ymax):
        while True:
            dep = Point(random.choice([-1, 0, 1]), random.choice([-1, 0, 1]))
            temp = self.pos + dep
            if dep != Point(0, 0) and temp != self.last_pos and (xmin <= temp.x <= xmax) and (ymin <= temp.y <= ymax):
                self.last_pos = self.pos
                self.pos = temp
                break


class Pacman(Entity):
    def move_to(self, ghost):
        self.pos += (ghost.pos - self.pos).sign()


class Pacman(Entity):
    def move_to(self, ghost):
        self.pos += (ghost.pos - self.pos).sign()

    def move(self, ghosts):
        nearest_ghost = None
        d2_min = 100000000
        for ghost in ghosts:
            d2 = (self.pos - ghost.pos).norm2()
            if d2 < d2_min:
                nearest_ghost = ghost
                d2_min = d2
        if nearest_ghost:
            self.move_to(nearest_ghost)
            if self.pos == nearest_ghost.pos:
                return nearest_ghost


class Ghost(Entity):
    pass


if __name__ == "__main__":
    # Test Entity move
    entity = Entity(4, 4, "TEST")
    print(entity)
    for i in range(20):
        entity.move(0, 0, 10, 5)
        print(entity)

    # Test Pacman move() method
    pacman = Pacman(0, 0, "PACMAN")
    print(pacman)
    for i in range(50):
        ghost = pacman.move([entity])
        print(pacman, "catched", ghost)

    # Test TeleportableGhost move() method  
    ghost = TeleportableGhost(0, 0, "GHOST")
    for i in range(50):
        ghost.move(0, 0, 10, 5)
        print(ghost)
    
